"""
Multi-label classification module for Python
============================================

Scikit-multilearn is a BSD-licensed library for multi-label classification that is
built on top of the well-known scikit-learn ecosystem.

"""
